package com.pseudocode.photolib.metadata;

import java.lang.reflect.Array;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.adobe.xmp.XMPConst;
import com.adobe.xmp.XMPDateTime;
import com.adobe.xmp.XMPException;
import com.adobe.xmp.XMPMeta;
import com.adobe.xmp.XMPUtils;
import com.adobe.xmp.options.PropertyOptions;
import com.adobe.xmp.properties.XMPProperty;
import com.pseudocode.photolib.utils.GPSUtil;
import com.pseudocode.photolib.utils.RationalUtil;

/**
 * XMPMeta to Meta factory
 */
public class MetaFactory {
	private static final Logger log = LoggerFactory.getLogger(MetaFactory.class);
	private static final DateFormat[] iso8601 = {
		new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSS", Locale.US),
		new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss", Locale.US),
		new SimpleDateFormat("yyyy-MM-dd'T'HH:mm", Locale.US),
		new SimpleDateFormat("yyyy-MM-dd", Locale.US)
	};

	private static interface XMPConverter<T> {
		T coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception;
	}

	/**
	 * Constructs a simplified Meta object from various representations within XMP metadata.
	 * @param xmp
	 * @return
	 */
	public static Meta create(XMPMeta xmp) {
		Meta meta = new Meta();
		if (xmp == null) {
			return meta;
		}

		try {
			meta.key(findString(xmp,
					XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_HASH));

			meta.name(findString(xmp,
					XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_NAME));

			meta.format(findString(xmp,
					XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_FORMAT));

			meta.size(findInteger(xmp,
					XMPUtil.NS_PHOTOLIB, XMPUtil.TAG_FILE_SIZE));

			meta.width(findInteger(xmp,
					XMPConst.NS_EXIF, "ExifImageWidth"));

			meta.height(findInteger(xmp,
					XMPConst.NS_EXIF, "ExifImageLength",
					XMPConst.NS_EXIF, "ExifImageHeight"));

			meta.timestamp(
				findDate(xmp,
					XMPConst.NS_EXIF, "DateTimeOriginal",
					XMPConst.NS_XMP, "CreateDate",
					XMPConst.NS_PHOTOSHOP, "DateCreated",
					XMPConst.NS_EXIF, "DateTimeDigitized",
					XMPConst.NS_EXIF, "DateTime",
					XMPConst.NS_EXIF, "GPSTimeStamp",
					XMPConst.NS_XMP, "ModifyDate",
					XMPConst.NS_PHOTOSHOP, "DigitalCreationDate"));

			meta.orientation(findInteger(xmp,
					XMPConst.NS_TIFF, "Orientation"));

			meta.latitude(findGPS(xmp,
					XMPConst.NS_EXIF, "GPSLatitude",
					XMPConst.NS_EXIF, "GPSDestLatitude"));

			meta.longitude(findGPS(xmp,
					XMPConst.NS_EXIF, "GPSLongitude",
					XMPConst.NS_EXIF, "GPSDestLongitude"));

			meta.compass(findRational(xmp,
					XMPConst.NS_EXIF, "GPSImgDirection"));

			meta.make(findString(xmp,
					XMPConst.NS_TIFF, "Make"));

			meta.model(findString(xmp,
					XMPConst.NS_TIFF, "Model"));

			meta.lens(findLens(xmp,
					XMPConst.NS_EXIF, "LensModel",
					XMPConst.NS_EXIF_AUX, "Lens",
					XMPConst.NS_EXIF_AUX, "LensInfo"));

			meta.focalLength(findRational(xmp,
					XMPConst.NS_EXIF, "FocalLength",
					XMPConst.NS_EXIF, "FocalLengthIn35mmFilm",
					XMPConst.NS_EXIF, "LensInfo"));

			meta.iso(findInteger(xmp,
					XMPConst.NS_EXIF, "ISO",
					XMPConst.NS_EXIF, "ISOSpeedRatings"));

			meta.fstop(findFStop(xmp,
					XMPConst.NS_EXIF, "FNumber",
					XMPConst.NS_EXIF, "ApertureValue"));

			meta.shutter(findShutter(xmp,
					XMPConst.NS_EXIF, "ExposureTime",
					XMPConst.NS_EXIF, "ShutterSpeedValue"));

			meta.exposureBias(findRationalString(xmp,
					XMPConst.NS_EXIF, "ExposureBiasValue",
					XMPConst.NS_EXIF, "ExposureCompensation"));

			meta.exposureMode(findInteger(xmp,
					XMPConst.NS_EXIF, "ExposureMode"));

			meta.flash(findInteger(xmp,
					XMPConst.NS_EXIF, "Flash"));

			meta.title(findString(xmp,
					XMPConst.NS_EXIF, "Title",
					XMPConst.NS_DC, "title",
					XMPConst.NS_TIFF, "XPTitle",
					XMPConst.NS_TIFF, "XPSubject"));

			meta.description(findString(xmp,
					XMPConst.NS_EXIF, "ImageDescription",
					XMPConst.NS_TIFF, "XPComment",
					XMPConst.NS_DC, "description"));

			meta.location(findString(xmp,
					XMPConst.NS_TIFF, "LocationCreated",
					XMPConst.NS_TIFF, "LocationShown"));

			meta.tags(findStringArray(xmp,
					XMPUtil.NS_LR, "hierarchicalSubject",
					XMPUtil.NS_MSPhoto, "LastKeywordXMP",
					XMPConst.NS_TIFF, "XPKeywords",
					XMPConst.NS_DC, "subject"));

			meta.creator(findString(xmp,
					XMPConst.NS_EXIF, "Artist",
					XMPConst.NS_DC, "creator",
					XMPConst.NS_PHOTOSHOP, "By-line",
					XMPConst.NS_XMP_RIGHTS, "UsageTerms"));

			meta.copyright(findString(xmp,
					XMPConst.NS_EXIF, "Copyright",
					XMPConst.NS_DC, "rights",
					XMPConst.NS_PHOTOSHOP, "CopyrightNotice",
					XMPConst.NS_XMP_RIGHTS, "UsageTerms"));

			meta.rating(findDouble(xmp,
					XMPConst.NS_XMP, "Rating"));

			/* tweaks -------------------------- */

			// Canon especially puts their brand in both fields
			String make = meta.make();
			String model = meta.model();
			if (make != null && model != null && model.startsWith(make)) {
				// trim the make off the start of model
				meta.model(model.substring(make.length()).trim());
			}

			// Sony just craps all over metadata
			if ("SONY DSC".equals(meta.description())) {
				meta.description(null);
			}

			// TODO: define a better differentiator. perhaps length or formatting?
			if ((meta.title() == null || meta.title().isEmpty()) &&
				(meta.description() != null && !meta.description().isEmpty())) {
				meta.title(meta.description());
				meta.description(null);
			}
			
		} catch (Exception ex) {
			log.error("metadata conversion error", ex);
		}
		return meta;
	}

	private static double findGPS(XMPMeta xmp, String... pairs) {
		Double value = find(xmp, new XMPConverter<Double>() {
				@Override
				public Double coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					return GPSUtil.parseDegrees((String)value);
				}
			},
			pairs);

		if (value == null) {
			return Double.NaN;
		}
		return value;
	}

	private static double findFStop(XMPMeta xmp, String... pairs) {
		Double value = find(xmp, new XMPConverter<Double>() {
				@Override
				public Double coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					int[] values = RationalUtil.parse((String)value);
					double doubleVal = RationalUtil.toDouble(values[0], values[1]);
					if ("ApertureValue".equals(name)) {
						// APEX Aperture Value (Av)
						// The actual aperture value of lens when the image was taken.
						// To convert this value to ordinary F-number(F-stop),
						// calculate this value's power of root 2 (=1.4142).
						// For example, if value is '5', F-number is 1.4142^5 = F5.6.
						return Math.pow(2.0, doubleVal/2.0);
					}
					return doubleVal;
				}
			},
			pairs);

		if (value == null) {
			return Double.NaN;
		}
		return value;
	}

	private static String findShutter(XMPMeta xmp, String... pairs) {
		return find(xmp, new XMPConverter<String>() {
				@Override
				public String coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					int[] values = RationalUtil.parse((String)value);
					if ("ShutterSpeedValue".equals(name)) {
						// APEX Shutter Speed Value (Tv)
						// To convert this value to ordinary 'Exposure Time';
						// calculate this value's power of 2, then reciprocal.
						// For example, if value is '4', shutter speed is 1/(2^4) = 1/16 second.
						if (values[0] > 0) {
							double speed = Math.pow(2.0, RationalUtil.toDouble(values[0], values[1]));
							return String.format("1/%d", (int)Math.round(speed));

						} else {
							double speed = Math.pow(2.0, -RationalUtil.toDouble(values[0], values[1]));
							return String.format("%.2f", speed);
						}
					}

					// ExposureTime rational
					if (values[0] < values[1]) {
						double speed = RationalUtil.toDouble(values[1], values[0]);
						return String.format("1/%d", (int)Math.round(speed));

					} else {
						double speed = RationalUtil.toDouble(values[0], values[1]);
						return String.format("%.2f", speed);
					}
				}
			},
			pairs);
	}

	private static String findLens(XMPMeta xmp, String... pairs) {
		return find(xmp, new XMPConverter<String>() {
				@Override
				public String coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					if ("FocalLength".equals(name) || "FocalLengthIn35mmFilm".equals(name)) {
						// parse and reduce the fraction
						int[] values = RationalUtil.parse((String)value);
						values = RationalUtil.reduce(values[0], values[1]);
						if (values[1] == 1) {
							return String.format("%dmm", values[0]);
						}
						return String.format("%.1fmm", RationalUtil.toDouble(values[0], values[1]));
					}
					if ("LensInfo".equals(name) && XMPConst.NS_EXIF_AUX.equals(ns)) {
						String result = null;
						String[] focalLengths = ((String)value).split("\\s+");
						for (int i=0; i<focalLengths.length; i++) {
							int[] values = RationalUtil.parse(focalLengths[i]);
							values = RationalUtil.reduce(values[0], values[1]);
							if (values[1] == 0) {
								focalLengths[i] = null;
								continue;
							}
							if (values[1] == 1) {
								focalLengths[i] = String.format("%d", values[0]);
							} else {
								focalLengths[i] = String.format("%.1f", RationalUtil.toDouble(values[0], values[1]));
							}
							if (result == null) {
								result = focalLengths[i];
							} else {
								result += "-"+focalLengths[i];
							}
						}
						return (result == null) ? null : result+"mm";
					}

					return (String)value;
				}
			},
			pairs);
	}

	private static String findRationalString(XMPMeta xmp, String... pairs) {
		String value = find(xmp, new XMPConverter<String>() {
				@Override
				public String coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					int[] values = RationalUtil.parse((String)value);
					values = RationalUtil.reduce(values[0], values[1]);
					return RationalUtil.toString(values[0], values[1]);
				}
			},
			pairs);

		return value;
	}

	private static double findRational(XMPMeta xmp, String... pairs) {
		Double value = find(xmp, new XMPConverter<Double>() {
				@Override
				public Double coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					int[] values = RationalUtil.parse((String)value);
					return RationalUtil.toDouble(values[0], values[1]);
				}
			},
			pairs);

		if (value == null) {
			return Double.NaN;
		}
		return value;
	}

	private static long findDate(XMPMeta xmp, String... pairs) {
		Long value = find(xmp, new XMPConverter<Long>() {
				@Override
				public Long coerce(XMPMeta xmp, String ns, String name, Object value) throws Exception {
					if (!(value instanceof String)) {
						return null;
					}

					try {
						XMPDateTime date = XMPUtils.convertToDate((String)value);
						return date.getCalendar().getTimeInMillis();
					} catch (XMPException ex) {
						log.warn("XMP Date error: "+value, ex);

						// check for dates without timezone
						for (DateFormat format : iso8601) {
							try {
								Date date = format.parse((String)value);
								// TODO: shift timezone?
								return date.getTime();
							} catch (Exception e) {}
						}
						return null;
					}
				}
			},
			pairs);

		if (value == null) {
			return 0L;
		}
		return value;
	}

	private static String findString(XMPMeta xmp, String... pairs) {
		return find(xmp, new XMPConverter<String>() {

			@Override
			public String coerce(XMPMeta xmp, String ns, String name, Object value) {
				if (value == null) {
					return null;
				}

				if (value.getClass().isArray()) {
					int length = Array.getLength(value);
					for (int i=0; i<length; i++) {
						Object item = Array.get(value, i);
						if (item == null) {
							continue;
						}
						String strVal = (item instanceof String) ? ((String)item).trim() : item.toString().trim();
						if (!strVal.isEmpty()) {
							return strVal;
						}
					}
					return null;
					
				} else {
					String strVal = (value instanceof String) ? ((String)value).trim() : value.toString().trim();
					return strVal.isEmpty() ? null : strVal;
				}
			}
			
		}, pairs);
	}

	private static String[] findStringArray(XMPMeta xmp, String... pairs) {
		return find(xmp, new XMPConverter<String[]>() {

			@Override
			public String[] coerce(XMPMeta xmp, String ns, String name, Object value) {
				if (value == null) {
					return null;
				}

				if (value.getClass().isArray()) {
					int length = Array.getLength(value);
					List<String> values = new ArrayList<String>(length);
					for (int i=0; i<length; i++) {
						Object item = Array.get(value, i);
						if (item == null) {
							continue;
						}
						String strVal = (item instanceof String) ? ((String)item).trim() : item.toString().trim();
						if (strVal.isEmpty()) {
							continue;
						}
						values.add(strVal);
					}
					if (values.size() < 1) {
						// avoid empty array being accepted
						return null;
					}
					return values.toArray(new String[values.size()]);
					
				} else {
					String strVal = (value instanceof String) ? ((String)value).trim() : value.toString().trim();
					return strVal.isEmpty() ? null : new String[] { strVal };
				}
			}
			
		}, pairs);
	}

	private static int findInteger(XMPMeta xmp, String... pairs) {
		Integer value = find(xmp, new XMPConverter<Integer>() {

				@Override
				public Integer coerce(XMPMeta xmp, String ns, String name, Object value) {
					if (value == null || value instanceof Integer) {
						return (Integer)value;
					}
					String strVal = (String)value;
					return strVal.isEmpty() ? null : Integer.parseInt(strVal);
				}

			}, pairs);

		if (value == null) {
			return 0;
		}
		return value;
	}

	private static double findDouble(XMPMeta xmp, String... pairs) {
		Double value = find(xmp, new XMPConverter<Double>() {

				@Override
				public Double coerce(XMPMeta xmp, String ns, String name, Object value) {
					if (value == null || value instanceof Double) {
						return (Double)value;
					}
					return Double.parseDouble((String)value);
				}
				
			}, pairs);

		if (value == null) {
			return Double.NaN;
		}
		return value;
	}

	/**
	 * Takes a list of XMP qualified name pairs and a conversion method, and
	 * returns the first successfully converted value.
	 * 
	 * @param xmp
	 * @param convertor
	 * @param pairs
	 * @return
	 */
	private static <T> T find(XMPMeta xmp, XMPConverter<T> convertor, String... pairs) {
		if (pairs == null || pairs.length < 1) {
			return null;
		}

		int length = pairs.length/2;
		for (int i=0; i<length; i++) {
			String ns = pairs[2*i];
			String name = pairs[2*i+1];
			if (!xmp.doesPropertyExist(ns, name)) {
				continue;
			}

			try {
				XMPProperty property = xmp.getProperty(ns, name);
				PropertyOptions options = property.getOptions();
				Object value;
				if (options.isArray() || options.isArrayOrdered() || options.isArrayAlternate() || options.isArrayAltText()) {
					int count = xmp.countArrayItems(ns, name);
					Object[] values = new Object[count];
					for (int index=0; index<count; index++) {
						values[index] = xmp.getArrayItem(ns, name, index+1).getValue();
					}
					value = (count == 1) ? values[0] : values;

				} else {
					value = xmp.getProperty(ns, name).getValue();
				}

				T result = convertor.coerce(xmp, ns, name, value);
				if (result != null) {
					return result;
				}
			} catch (Exception ex) {
				log.warn("metadata extraction error", ex);
			}
		}
		return null;
	}
}
