package com.pseudocode.photolib.utils;

/**
 * GPS coordinate utilities
 */
public class GPSUtil {

	public final static char NORTH = 'N';	// positive
	public final static char EAST = 'E';	// positive
	public final static char WEST = 'W';	// negative
	public final static char SOUTH = 'S';	// negative

	private final static double SEC_PER_MIN = 60.0;
	private final static double MIN_PER_DEG = 60.0;

	/**
	 * Splits a single degrees field (DegDec) into degrees, minutes (MinDec)
	 * @param value
	 * @return [degrees, minutes]
	 */
	public static double[] toDegreesMinutes(double value) {
		// store sign, work in positive space
		double sign = Math.signum(value);
		value = Math.abs(value);

		// split out into individual fields
		double deg = sign * Math.floor(value);
		double min = (value % 1.0) * MIN_PER_DEG;

		return new double[] {
			deg,
			min
		};
	}

	/**
	 * Splits a single degrees field (DegDec) into degrees, minutes, seconds (DMS)
	 * @param value
	 * @return [degrees, minutes, seconds]
	 */
	public static double[] toDegreesMinutesSeconds(double value) {
		// store sign, work in positive space
		double sign = Math.signum(value);
		value = Math.abs(value);

		// split out into individual fields
		double degrees = sign * Math.floor(value);
		double minutes = (value % 1.0) * MIN_PER_DEG;
		double seconds = (minutes % 1.0) * SEC_PER_MIN;
		minutes = Math.floor(minutes);

		return new double[] {
			degrees,
			minutes,
			seconds
		};
	}

	/**
	 * Combines degrees, minutes, seconds (DMS) into a single degrees field (DegDec) 
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @return degrees
	 */
	public static double toDegrees(double degrees, double minutes) {
		// store sign, work in positive space
		double sign = Math.signum(degrees);
		degrees = Math.abs(degrees);

		// recombine
		return sign * (degrees + (minutes / MIN_PER_DEG));
	}

	/**
	 * Combines degrees, minutes, seconds (DMS) into a single degrees field (DegDec) 
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @return degrees
	 */
	public static double toDegrees(double degrees, double minutes, double seconds) {
		// store sign, work in positive space
		double sign = Math.signum(degrees);
		degrees = Math.abs(degrees);

		// recombine
		return sign * (degrees + ((minutes + (seconds / SEC_PER_MIN)) / MIN_PER_DEG));
	}

	/**
	 * Combines degrees, minutes, seconds (DMS) into a single degrees field (DegDec) 
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @param direction directional indicator N/E/W/S
	 * @return degrees
	 */
	public static double toDegrees(double degrees, double minutes, double seconds, char direction) {
		int dir;
		switch (Character.toUpperCase(direction)) {
			case WEST:
			case SOUTH:
				dir = -1;
				break;
			default:
				dir = 1;
				break;
		}

		return toDegrees(dir*Math.abs(degrees), minutes, seconds);
	}

	/**
	 * Combines degrees, minutes, seconds (DMS) into a single degrees field (DegDec) 
	 * @param dms [degrees, minutes, seconds]
	 * @return degrees
	 */
	public static double toDegrees(double[] dms) {
		if (dms == null) {
			throw new NullPointerException("dms");
		}

		// ensure individual fields
		double degrees = (dms.length > 0) ? dms[0] : 0.0;
		double minutes = (dms.length > 1) ? dms[1] : 0.0;
		double seconds = (dms.length > 2) ? dms[2] : 0.0;

		// store sign, work in positive space
		double sign = Math.signum(degrees);
		degrees = Math.abs(degrees);

		// recombine
		return sign * (degrees + ((minutes + (seconds/SEC_PER_MIN)) / MIN_PER_DEG));
	}

	/**
	 * Parses a string containing DMS, XMP, numeric value
	 * @param value
	 * @return
	 */
	public static double parseDegrees(String value) {
		if (value == null || value.isEmpty()) {
			throw new NullPointerException("value");
		}

		boolean negate = false;
		double[] dms = new double[3];
		int unit = 0,
			index = 0,
			start = 0;

		for (; index<value.length(); index++) {
			char ch = value.charAt(index);

			if (ch == '.' ||
				ch == '-' ||
				ch == '+' ||
				(ch >= '0' && ch <= '9')) {
				// accumulate numeric chars
				continue;
			}

			if (start < index && unit < dms.length) {
				dms[unit] = Double.parseDouble(value.substring(start, index));
				unit++;
			}

			switch (ch) {
				case NORTH:
				case EAST:
					negate = false;
					break;
				case WEST:
				case SOUTH:
					negate = true;
					break;
			}
			start = index+1;
		}

		if (start < index && unit < dms.length) {
			dms[unit] = Double.parseDouble(value.substring(start, index));
			unit++;
		}

		if (negate) {
			dms[0] = -dms[0];
		}

		return toDegrees(dms);
	}

	/**
	 * Formats as exif:GPSLatitude XMP-style string, e.g. DD,MM.mmk
	 * @param latitude
	 * @return
	 */
	public static String toXMPStringLatitude(double latitude) {
		double[] dms = toDegreesMinutesSeconds(latitude);
		char direction = (dms[0] < 0) ? SOUTH : NORTH;

		return toXMPString((int)dms[0], (int)dms[1], dms[2], direction);
	}

	/**
	 * Formats as exif:GPSLongitude XMP-style string, e.g., "DD,MM.mmk"
	 * @param longitude
	 * @return
	 */
	public static String toXMPStringLongitude(double longitude) {
		double[] dms = toDegreesMinutesSeconds(longitude);
		char direction = (dms[0] < 0) ? WEST : EAST;

		return toXMPString((int)dms[0], (int)dms[1], dms[2], direction);
	}

	/**
	 * Formats as XMP-style string, e.g., "DD,MM,SSk" or "DD,MM.mmk"
	 * @param degrees
	 * @param minutes
	 * @param direction
	 * @return
	 */
	public static String toXMPString(int degrees, int minutes, double seconds, char direction) {
		double precision = seconds - Math.floor(seconds);

		if (precision < 1e-6) {
			if ((int)seconds == 0) {
				if (minutes == 0) {
					return String.format(
						"%d%S",
						Math.abs(degrees),
						direction);
				}

				return String.format(
					"%d,%d%S",
					Math.abs(degrees),
					minutes,
					direction);
			}

			return String.format(
				"%d,%d,%d%S",
				Math.abs(degrees),
				minutes,
				(int)seconds,
				direction);
		}

		return String.format(
			"%d,%.6f%S",
			Math.abs(degrees),
			minutes+(seconds / SEC_PER_MIN),
			direction);
	}

	/**
	 * Formats latitude as human-readable degrees, minutes, seconds. e.g., "DD°MM′SS.sss″k"
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @param direction
	 * @return
	 */
	public static String toDMSStringLatitude(double latitude) {
		double[] dms = toDegreesMinutesSeconds(latitude);
		char direction = (dms[0] < 0) ? SOUTH : NORTH;

		return toDMSString((int)dms[0], (int)dms[1], dms[2], direction);
	}

	/**
	 * Formats longitude as human-readable degrees, minutes, seconds. e.g., "DD°MM′SS.sss″k"
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @param direction
	 * @return
	 */
	public static String toDMSStringLongitude(double longitude) {
		double[] dms = toDegreesMinutesSeconds(longitude);
		char direction = (dms[0] < 0) ? WEST : EAST;

		return toDMSString((int)dms[0], (int)dms[1], dms[2], direction);
	}

	/**
	 * Formats as human-readable degrees, minutes, seconds. e.g., "DD°MM′SS.sss″k"
	 * @param degrees
	 * @param minutes
	 * @param seconds
	 * @param direction
	 * @return
	 */
	public static String toDMSString(int degrees, int minutes, double seconds, char direction) {
		double precision = seconds - Math.floor(seconds);

		if (precision < 1e-6) {
			if ((int)seconds == 0) {
				if (minutes == 0) {
					return String.format(
							"%d\u00B0%S",
							Math.abs(degrees),
							direction);
				}

				return String.format(
						"%d\u00B0%d\u2032%S",
						Math.abs(degrees),
						minutes,
						direction);
			}

			return String.format(
				"%d\u00B0%d\u2032%d\u2033%S",
				Math.abs(degrees),
				minutes,
				(int)seconds,
				direction);

		}

		return String.format(
			"%d\u00B0%d\u2032%.6f\u2033%S",
			Math.abs(degrees),
			minutes,
			seconds,
			direction);
	}
}
