package com.pseudocode.photolib.utils;

import java.util.Arrays;

/*
 * Implements a fast 64-bit binary version of Geohash algorithm
 * 		http://en.wikipedia.org/wiki/Geohash
 * 		http://geohash.org
 */
public class GeohashUtil {
	// the Base32 alphabet used by geohash.org is different from RFC-4648
	private static final String BASE32 = "0123456789bcdefghjkmnpqrstuvwxyz";
	private static final int BASE32_BITS = getBitsPerChar(BASE32.length());
	private static final int LON_POS = 0;
	private static final int LAT_POS = 1;
	private static final int LAT_LON_DIM = 2;
	private static final double LAT_MIN = -90.0;
	private static final double LAT_MAX = 90.0;
	private static final double LON_MIN = -180.0;
	private static final double LON_MAX = 180.0;
	private static final double EARTH_RADIUS = 6371.009;// kilometers
	private static final double TO_RADIANS = Math.PI / 180.0;

	public static String[] getQueryRanges(String hash, int precision) {
		long[] prefixes = getQueryRanges(decodeLong(hash, BASE32, BASE32_BITS), precision);
		String[] hashes = new String[prefixes.length];
		for (int i=0; i<prefixes.length; i++) {
			hashes[i] = encodeLong(prefixes[i], BASE32, BASE32_BITS);
		}
		return hashes;
	}

	/**
	 * Gets query range values to search
	 * @param geohash of query center point
	 * @param bits of precision (1-30)
	 * @return pairs of inclusive ranges
	 */
	public static long[] getQueryRanges(long hash, int precision) {
		return getQueryRanges(hash, precision, true);
	}
	
	/**
	 * Gets query range values to search
	 * @param geohash of query center point
	 * @param bits of precision (1-30)
	 * @return pairs of inclusive ranges
	 */
	public static long[] getQueryRanges(long hash, int precision, boolean coalesce) {
		final int SOUTH_POLE = 0;
		final int NORTH_POLE = -1;

		if (precision <= 0 || precision >= Integer.SIZE) {
			throw new IllegalArgumentException("precision");
		}

		// prefix mask and minimum feature size (region)
		final int inc = 1 << (Integer.SIZE - precision);
		final int mask = -inc;

		// split component values from hash
		final int lat = deinterlaceMorton(hash);
		final int lon = deinterlaceMorton(hash >>> 1);

		// determine which quadrant the point sits to determine neighboring regions
		// if is in upper quadrant effectively shift to upper edge to get upper neighbors
		final boolean shiftLat = ((lat & (inc >>> 1)) != 0) && ((lat & mask) >= 0 || (lat & mask) < -inc);
		final boolean shiftLon = ((lon & (inc >>> 1)) != 0);

		// mask off prefixes aligning to lower edge of region
		// if was in upper quadrant, shift up to next region
		// this point then becomes the center vertex of the 2x2 search grid
		final int latC = shiftLat ? (lat & mask) + inc : (lat & mask);
		final int lonC = shiftLon ? (lon & mask) + inc : (lon & mask);

		// get the lower and upper grid vertices
		// must prevent overflow/underflow for latitude
		// must allow underflow for longitude (means crossing international date-line)
		final int latL = (latC >= 0 && latC < inc) ? SOUTH_POLE : (latC - inc);
		final int latH = (latC < 0 && latC >= -inc) ? NORTH_POLE : (latC + inc);
		final int lonL = (lonC - inc);
		final int lonH = (lonC + inc);

		long[] ranges;

		// listed in (unsigned long) ascending order
		if (latL == latC) {
			ranges = new long[] {

				// W region
				interlaceMorton(latL) | (interlaceMorton(lonL) << 1),
				interlaceMorton(latH-1) | (interlaceMorton(lonC-1) << 1),

				// E region
				interlaceMorton(latL) | (interlaceMorton(lonC) << 1),
				interlaceMorton(latH-1) | (interlaceMorton(lonH-1) << 1)
			};

		} else {
			ranges = new long[] {

				// SW region
				interlaceMorton(latL) | (interlaceMorton(lonL) << 1),
				interlaceMorton(latC-1) | (interlaceMorton(lonC-1) << 1),
	
				// NW region
				interlaceMorton(latC) | (interlaceMorton(lonL) << 1),
				interlaceMorton(latH == NORTH_POLE ? NORTH_POLE : latH-1) | (interlaceMorton(lonC-1) << 1),
	
				// SE region
				interlaceMorton(latL) | (interlaceMorton(lonC) << 1),
				interlaceMorton(latC-1) | (interlaceMorton(lonH-1) << 1),
	
				// NE region
				interlaceMorton(latC) | (interlaceMorton(lonC) << 1),
				interlaceMorton(latH == NORTH_POLE ? NORTH_POLE : latH-1) | (interlaceMorton(lonH-1) << 1)
			};
		}

		return coalesce ? coalesceRanges(ranges) : ranges;
	}

	static long[] coalesceRanges(long[] ranges) {
		// combine any contiguous ranges
		int count = ranges.length;
		for (int i=1; i+1<count; i+=2) {
			if (Math.abs(ranges[i] - ranges[i+1]) == 1) {
				// remove adjacent end points, shift remaining elements
				for (int j=i; j+2<count; j++) {
					ranges[j] = ranges[j+2];
				}
				count -= 2;
				i -= 2;
			}
		}

		if (count == ranges.length) {
			// no change
			return ranges;
		}

		// remove any unused elements
		return Arrays.copyOf(ranges, count);
	}

	/**
	 * Encodes latitude/longitude to a 64-bit binary representation of Geohash
	 * @param latitude
	 * @param longitude
	 * @return hash
	 */
	public static long encode(double latitude, double longitude) {
		
		// interlace encoded Morton code bits
		return
			(interlaceMorton(encodeZOrder(latitude, LAT_MIN, LAT_MAX))) |
			(interlaceMorton(encodeZOrder(longitude, LON_MIN, LON_MAX)) << 1);
	}

	/**
	 * Decodes latitude/longitude from a 64-bit binary representation of Geohash
	 * @param hash
	 * @return point [latitude, longitude]
	 */
	public static double[] decode(long hash) {
		// decode deinterlaced bits of Morton code 
		return new double[] {
			(decodeZOrder(deinterlaceMorton(hash), LAT_MIN, LAT_MAX)),
			(decodeZOrder(deinterlaceMorton(hash >>> 1), LON_MIN, LON_MAX))
		};
	}

	/**
	 * Encodes a single dimension of Z-order curve
	 * @param val
	 * @param min lower bound
	 * @param max upper bound
	 * @return output
	 */
	private static int encodeZOrder(double val, double min, double max) {
		int hash = 0;

		for (int bits=Integer.SIZE; bits>0; bits--) {
			hash <<= 1;
			double mid = (min + max) / 2.0;
			if (val > mid) {
				min = mid;
				hash |= 1;
			} else {
				max = mid;
			}
		}

		return hash;
	}

	/**
	 * Decodes a single dimension of Z-order curve
	 * @param hash Z-Order hash
	 * @param min lower bound
	 * @param max upper bound
	 */
	private static double decodeZOrder(int hash, double min, double max) {

		int mask = 1 << (Integer.SIZE-1);

		for (int bits=Integer.SIZE; bits>0; bits--) {
			double mid = (min + max) / 2.0;
			if ((hash & mask) != 0L) {
				min = mid;
			} else {
				max = mid;
			}
			mask >>>= 1;
		}

		return (min + max) / 2.0;
	}

	/**
	 * Expands 32-bits up to 64-bits by inserting 0 bits after each bit
	 * @param x
	 * @return
	 */
	private static long interlaceMorton(int x) {
		long z = x & 0x00000000FFFFFFFFL;
		z = (z | (z << 16)) & 0x0000FFFF0000FFFFL;
		z = (z | (z <<  8)) & 0x00FF00FF00FF00FFL;
		z = (z | (z <<  4)) & 0x0F0F0F0F0F0F0F0FL;
		z = (z | (z <<  2)) & 0x3333333333333333L;
		z = (z | (z <<  1)) & 0x5555555555555555L;
		return z;
	}

	/**
	 * Compacts 64-bits down to 32-bits by removing all odd bits and compacting
	 * @param z
	 * @return
	 */
	private static int deinterlaceMorton(long z) {
		z &= 0x5555555555555555L;
		z = (z | (z >>>  1)) & 0x3333333333333333L;
		z = (z | (z >>>  2)) & 0x0F0F0F0F0F0F0F0FL;
		z = (z | (z >>>  4)) & 0x00FF00FF00FF00FFL;
		z = (z | (z >>>  8)) & 0x0000FFFF0000FFFFL;
		z = (z | (z >>> 16)) & 0x00000000FFFFFFFFL;
		return (int)z;
	}

	/**
	 * Encodes latitude/longitude to a 64-bit binary representation of Geohash (alternate implementation)
	 * @param latitude
	 * @param longitude
	 * @return hash
	 */
	public static long encode2(double latitude, double longitude) {
		// interlace encoded Morton code bits
		return
			encodeZMorton(latitude, LAT_MIN, LAT_MAX, LAT_POS, LAT_LON_DIM) |
			encodeZMorton(longitude, LON_MIN, LON_MAX, LON_POS, LAT_LON_DIM);
	}

	/**
	 * Decodes latitude/longitude from a 64-bit binary representation of Geohash (alternate implementation)
	 * @param hash
	 * @return point [latitude, longitude]
	 */
	public static double[] decode2(long hash) {
		// decode deinterlaced bits of Morton code 
		return new double[] {
			decodeZMorton(hash, LAT_MIN, LAT_MAX, LAT_POS, LAT_LON_DIM),
			decodeZMorton(hash, LON_MIN, LON_MAX, LON_POS, LAT_LON_DIM)
		};
	}

	/**
	 * Encodes a single dimension of 64-bit Morton code
	 * @param val
	 * @param min lower bound
	 * @param max upper bound
	 * @param pos Morton code dimension position
	 * @param dim number of Morton code dimensions
	 * @return output
	 */
	private static long encodeZMorton(double val, double min, double max, int pos, int dim) {
		long hash = 0L;

		for (int bits=(Long.SIZE/dim); bits>0; bits--) {
			hash <<= dim;
			double mid = (min + max) / 2.0;
			if (val > mid) {
				min = mid;
				hash |= 1L;
			} else {
				max = mid;
			}
		}

		return (hash << (dim-pos-1));
	}

	/**
	 * Decodes a single dimension of 64-bit Morton code
	 * @param hash Morton code
	 * @param min lower bound
	 * @param max upper bound
	 * @param pos Morton code dimension position
	 * @param dim number of Morton code dimensions
	 */
	private static double decodeZMorton(long hash, double min, double max, int pos, int dim) {

		long mask = 1L << (Long.SIZE-1);
		mask >>>= pos;

		for (int bits=(Long.SIZE/dim); bits>0; bits--) {
			double mid = (min + max) / 2.0;
			if ((hash & mask) != 0L) {
				min = mid;
			} else {
				max = mid;
			}
			mask >>>= dim;
		}

		return (min + max) / 2.0;
	}

	/**
	 * Encodes latitude/longitude to a Base32 string representation of Geohash
	 * @param latitude
	 * @param longitude
	 * @return hash
	 */
	public static String encodeHash(double latitude, double longitude) {
		// interlace encoded Morton code bits
		return encodeLong(
			encode(latitude, longitude),
			BASE32,
			BASE32_BITS
		);
	}

	/**
	 * Decodes latitude/longitude from a Base32 string representation of Geohash
	 * @param hash
	 * @return point [latitude, longitude]
	 */
	public static double[] decodeHash(String hash) {
		if (hash == null) {
			throw new NullPointerException("hash");
		}

		// decode deinterlaced bits of Morton code 
		return decode(
			decodeLong(
				hash,
				BASE32,
				BASE32_BITS)
		);
	}

	/**
	 * Calculate the distance between two Geohash values
	 * @param hash1 Geohash 1
	 * @param hash2 Geohash 2
	 * @return kilometers
	 */
	public static double distance(String hash1, String hash2) {
		return distance(
			decodeHash(hash1),
			decodeHash(hash2));
	}

	/**
	 * Calculate the distance between two Geohash values
	 * @param hash1 Geohash 1
	 * @param hash2 Geohash 2
	 * @return kilometers
	 */
	public static double distance(long hash1, long hash2) {
		return distance(
			decode(hash1),
			decode(hash2));
	}

	/**
	 * Calculate the distance between two latitude-longitude points
	 * @param pt1 point 1 [latitude, longitude]
	 * @param pt2 point 2 [latitude, longitude]
	 * @return kilometers
	 */
	public static double distance(double[] pt1, double[] pt2) {
		return distance(
			pt1[0], pt1[1],
			pt2[0], pt2[1]);
	}

	/**
	 * Calculate the distance between two latitude-longitude points
	 * @param lat1 point 1 latitude
	 * @param lng1 point 1 longitude
	 * @param lat2 point 2 latitude
	 * @param lng2 point 2 longitude
	 * @return kilometers
	 */
	public static double distance(double lat1, double lng1, double lat2, double lng2) {
		// Implements the haversine formula for great-circle distance
		//		http://en.wikipedia.org/wiki/Haversine_formula
		//		http://en.wikipedia.org/wiki/Great-circle_distance

		double dLat = (lat2-lat1) * TO_RADIANS;
		double dLng = (lng2-lng1) * TO_RADIANS;

		double aLat = Math.sin(dLat / 2.0);
		double aLng = Math.sin(dLng / 2.0);
		aLat *= aLat; //squared
		aLng *= aLng; //squared

		double a = aLat + Math.cos(lat1 * TO_RADIANS) * Math.cos(lat2 * TO_RADIANS) * aLng;
		double c = 2.0 * Math.atan2(Math.sqrt(a), Math.sqrt(1.0-a));

		return c * EARTH_RADIUS;
	}

	/**
	 * Encodes latitude/longitude to a Base32 string representation of Geohash
	 * @param latitude
	 * @param longitude
	 * @return hash
	 */
	public static String encodeLong(long hash) {
		return encodeLong(
			hash,
			BASE32,
			BASE32_BITS
		);
	}

	/**
	 * Encodes the value using an arbitrary alphabet
	 * @param value
	 * @param alphabet
	 * @param bitsPerChar
	 * @return
	 */
	private static String encodeLong(long value, String alphabet, int bitsPerChar) {
		final int count = Long.SIZE;
		StringBuilder result = new StringBuilder((int) Math.ceil((double)count / (double)bitsPerChar));

		long mask = 1L << (count-1);

		int i = 0, digit = 0;
		while (i < count) {
			digit <<= 1;
			if ((value & mask) != 0L) {
				digit |= 1;
			}
			mask >>>= 1;
			i++;
			if (i % bitsPerChar == 0) {
				result.append(alphabet.charAt(digit));
				digit = 0;
			}
		}

		return result.toString();
	}

	/**
	 * Decodes binary geohash from a Base32 string representation of Geohash
	 * @param hash
	 * @return binary geohash
	 */
	public static long decodeLong(String hash) {
		if (hash == null) {
			throw new NullPointerException("hash");
		}

		// decode deinterlaced bits of Morton code 
		return decodeLong(
			hash,
			BASE32,
			BASE32_BITS);
	}

	/**
	 * Decodes the hash using an arbitrary alphabet
	 * @param value
	 * @param alphabet
	 * @param bitsPerChar
	 * @return
	 */
	private static long decodeLong(String value, String alphabet, int bitsPerChar) {
		final int max = (int) Math.ceil((double)Long.SIZE / (double)bitsPerChar);
		final int count = value.length();
		final int overflow = (max * bitsPerChar) - Long.SIZE;

		long result = 0L;
		for (int i=0; i<max; i++) {
			int digit = (i < count) ? alphabet.indexOf(value.charAt(i)) : 0;
			if (digit < 0) {
				throw new IllegalArgumentException("illegal hash chars: "+value);
			}

			if (i == max-1 && overflow != 0) {
				result <<= bitsPerChar - overflow;
				if (overflow > 0) {
					// overflow, truncate less significant digit bits
					digit >>>= overflow;
				} else {
					// underflow, pad last digit with zeros
					digit <<= -overflow;
				}
			} else {
				result <<= bitsPerChar;
			}
			result |= digit;
		}

		return result;
	}

	/**
	 * Determines the number of bits per digit of the alphabet
	 * @param alphabetSize
	 * @return number of bits
	 */
	private static int getBitsPerChar(int alphabetSize) {
		// Math.pow(2, bitsPerChar) == alphabet.length()
		return (int) Math.floor( Math.log(alphabetSize) / Math.log(2));
	}
}
