package com.pseudocode.photolib.utils;

/**
 * Rational number utilities
 */
public class RationalUtil {

	private static final char DELIM = '/';

	/**
	 * Approximate the double value accurate to a precision of 0.000001
	 * http://stackoverflow.com/questions/95727
	 * 
	 * @param value double value to approximate
	 * @return an approximation of the value as a rational number
	 */
	public static int[] approximate(double value) {
		return approximate(value, 0.000001);
	}

	/**
	 * Approximate the double value accurate to a precision of 0.000001
	 * http://stackoverflow.com/questions/95727
	 * 
	 * @param value double value to approximate
	 * @param epsilon maximum precision to converge
	 * @return an approximation of the value as a rational number
	 */
	public static int[] approximate(double value, double epsilon) {
		if (value <= Integer.MIN_VALUE) {
			return new int[] {
				Integer.MIN_VALUE,
				1
			};
		}
		if (value >= Integer.MAX_VALUE) {
			return new int[] {
				Integer.MAX_VALUE,
				1
			};
		}

		// store sign and work in positive space
		int sign = (int)Math.signum(value);
		value = Math.abs(value);

		int numerator = (int)Math.round(value);
		int denominator = 1;
		double fraction = (double)numerator;// denominator is 1

		// continue until close enough or reached max integer
		while ((Math.abs(fraction-value) > epsilon) &&
				(denominator < Integer.MAX_VALUE) &&
				(numerator < Integer.MAX_VALUE)) {

			if (fraction < value) {
				numerator++;

			} else {
				denominator++;

				// calc new numerator
				long temp = Math.round(value * denominator);

				// test for overflow
				if (temp > Integer.MAX_VALUE) {
					denominator--;
					break;
				}

				numerator = (int)temp;
			}

			fraction = (double)numerator / (double)denominator;
		}

		return new int[] {
			sign*numerator,
			denominator
		};
	}

	/**
	 * Converts the String representation of a number to its rational equivalent.
	 * 
	 * @param value double value to approximate
	 * @return an approximation of the value as a rational number
	 */
	public static int[] parse(String value) {
		if (value == null || value.isEmpty()) {
			return new int[]{ 0, 0 };
		}

		int slash = value.indexOf(DELIM);
		if (slash < 0) {
			// approximate as floating point
			return approximate(Double.parseDouble(value.trim()));
		}

		return new int[] {
			Integer.parseInt(value.substring(0, slash).trim()),
			Integer.parseInt(value.substring(slash+1).trim())
		};
	}

	/**
	 * Finds the greatest common divisor and reduces the fraction by this amount.
	 * 
	 * @return the reduced rational
	 */
	public static int[] reduce(int numerator, int denominator) {

		// greatest common divisor
		int gcd = gcd(numerator, denominator);
		if (gcd != 1 && gcd != 0) {
			numerator /= gcd;
			denominator /= gcd;
		}

		// normalize signs
		if (denominator < 0) {
			numerator = -numerator;
			denominator = -denominator;
		}

		return new int[] {
			numerator,
			denominator
		};
	}

	public static int[] add(int n1, int d1, int n2, int d2) {
		int denominator = lcd(d1, d2);
		if (denominator > d1) {
			n1 *= (denominator/d1);
		}
		if (denominator > d2) {
			n2 *= (denominator/d2);
		}

		int numerator = n1 + n2;

		return new int[] {
			numerator,
			denominator
		};
	}

	public static int[] subtract(int n1, int d1, int n2, int d2) {
		return add(n1, d1, -n2, d2);
	}

	public static int[] multiply(int n1, int d1, int n2, int d2) {
		return reduce(n1*n2, d1*d2);
	}

	public static int[] divide(int n1, int d1, int n2, int d2) {
		return multiply(n1, d1, d2, n2);
	}
	
	/**
	 * Lowest Common Denominator
	 * @param a
	 * @param b
	 * @return lcd
	 */
	private static int lcd(int a, int b) {
		if (a == 0 && b == 0) {
			return 0;
		}

		return (a * b) / gcd(a, b);
	}

	/**
	 * Greatest Common Divisor
	 * @param a
	 * @param b
	 * @return gcd
	 */
	private static int gcd(int a, int b) {
		if (a < 0) {
			a = -a;
		}
		if (b < 0) {
			b = -b;
		}

		while (a != b) {
			if (a == 0) {
				return b;
			}
			if (b == 0) {
				return a;
			}

			if (a > b) {
				a %= b;

			} else {
				b %= a;
			}
		}
		return a;
	}

	public static double toDouble(int numerator, int denominator) {
		if (denominator == 0) {
			return 0.0;
		}

		return (double)numerator / (double)denominator;
	}

	public static String toString(int numerator, int denominator) {
		// normalize signs
		if (denominator < 0) {
			numerator = -numerator;
			denominator = -denominator;
		}

		if (denominator == 1) {
			return Integer.toString(numerator);
		}
		return Integer.toString(numerator)+DELIM+Integer.toString(denominator);
	}
}
